<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 10/02/2012 *
 * ===================                          Last change: 10/02/2012 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : install-page_finalize.php                        *
 * -------------------------------------------------------------------- *
 * Short description : Installation page include file                   *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Include-Datei fuer Installationsseite            *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2013 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if ((!defined('__SECURITY')) || (!isInstaller())) {
	die();
} // END - if

// Check pre-conditions
if ((isPostRequestElementSet('finalize')) && (!isInstalled())) {
	// You have submitted data then we have to reset the SQLs
	initSqls();

	// Remove any existing "cache"
	unsetSqlLinkUp(__FILE__, __LINE__);

	// Connect to database server
	sqlConnectToDatabase($GLOBALS['install_mysql']['host'], $GLOBALS['install_mysql']['login'], $GLOBALS['install_mysql']['password1'], __FILE__, __LINE__);

	// Is the link up?
	if (isSqlLinkUp()) {
		// Seems to work, also right database?
		if (sqlSelectDatabase($GLOBALS['install_mysql']['dbase'], __FILE__, __LINE__) === TRUE) {
			// Check for dumps
			if (!isInstallerSqlsReadable(postRequestElement('spath'))) {
				// Installation area not found
				reportBug(__FILE__, __LINE__, 'SQL dumps not found. Please extract ALL files from the archive or checkout all files out from SVN.');
				return;
			} // END - if

			// Any errors detected?
			if (!ifFatalErrorsDetected()) {
				// Set type, prefix from POST data and database name for later queries
				setConfigEntry('_TABLE_TYPE'  , postRequestElement('mysql', 'type'));
				setConfigEntry('_MYSQL_PREFIX', postRequestElement('mysql', 'prefix'));
				setConfigEntry('__DB_NAME'    , $GLOBALS['install_mysql']['dbase']);

				// Set path
				setConfigEntry('PATH', postRequestElement('spath'));
				unset($GLOBALS['getPath']);

				// Both exists so import them
				importInstallSqlDump('tables');
				importInstallSqlDump('menu-' . getLanguage());
				//* DEBUG: */ die(__FUNCTION__.'['.__LINE__.']:'<pre>'.print_r(getSqls(), TRUE).'</pre>');

				// Are some SQLs found?
				if (countSqls() == 0) {
					// Abort here
					addFatalMessage(__FILE__, __LINE__, '{--INSTALLER_SQL_IMPORT_FAILED--}');
					return;
				} // END - if

				// Now run all queries through
				runFilterChain('run_sqls');
			} // END - if
		} // END - if
	} // END - if

	// Are some fatal errors there?
	if (ifFatalErrorsDetected()) {
		$OUT = '';
		foreach (getFatalArray() as $value) {
			$OUT .= '    <li>' . $value . '</li>';
		} // END foreach
		$content['fatal_errors'] = $OUT;
		$OUT = '';
		foreach ($GLOBALS['install_mysql'] as $key => $value) {
			$OUT .= '    <input type="hidden" name="mysql[' . $key . ']" value="' . $value . '" />';
		} // END foreach
		$content['mysql_hidden']  = $OUT;
		$content['spath']         = postRequestElement('spath');
		$content['burl']          = postRequestElement('burl');
		$content['title']         = postRequestElement('title');
		$content['slogan']        = postRequestElement('slogan');
		$content['email']         = postRequestElement('email');
		$content['smtp_host']     = postRequestElement('smtp_host');
		$content['smtp_user']     = postRequestElement('smtp_user');
		$content['smtp_password'] = postRequestElement('smtp_password1');

		// Load template
		addTemplateToInstallContent('install_fatal_errors', $content);

		// We have handled all fatal errors here
		initFatalMessages();
	} else {
		// Copy the config template and verify it
		doInstallWriteLocalConfigurationFile(
			postRequestElement('spath'),
			postRequestElement('burl'),
			postRequestElement('title'),
			postRequestElement('slogan'),
			postRequestElement('email'),
			postRequestElement('warn_no_pass'),
			postRequestElement('wfooter'),
			postRequestElement('blink'),
			postRequestElement('mysql', 'host'),
			postRequestElement('mysql', 'dbase'),
			postRequestElement('mysql', 'login'),
			postRequestElement('mysql', 'password1'),
			postRequestElement('mysql', 'prefix'),
			postRequestElement('mysql', 'type'),
			postRequestElement('mysql', 'extension'),
			postRequestElement('smtp_host'),
			postRequestElement('smtp_user'),
			postRequestElement('smtp_password1')
		);

		// Register ext-sql_patches and ext-task
		if ((registerExtension('sql_patches', NULL)) && (registerExtension('task', NULL))) {
			// Installation is done!
			redirectToUrl('install.php?install_page=finished');
		} else {
			// Something goes wrong on registration of ext-sql_patches
			addFatalMessage(__FILE__, __LINE__, '{--INSTALLER_FINALIZER_CANNOT_REGISTER_SQL_PATCHES--}');
		}
	}
} else {
	// Something goes wrong during installation! :-(
	addFatalMessage(__FILE__, __LINE__, '{--INSTALLER_FINALIZER_FAILED--}');
}

// [EOF]
?>
