<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 09/09/2008 *
 * ===================                          Last change: 09/09/2008 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : load-config.php                                  *
 * -------------------------------------------------------------------- *
 * Short description : Load more cache files                            *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Mehr Cache-Dateien nachladen                     *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2013 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} elseif (isInstaller()) {
	// Do not run in installation phase
	//* DEBUG: */ logDebugMessage(__FILE__, __LINE__, 'Run from installation phase.');
	return FALSE;
}

// Let's start with the admins table...
if (($GLOBALS['cache_instance']->loadCacheFile('admin')) && ($GLOBALS['cache_instance']->extensionVersionMatches('admins'))) {
	// Load cache
	$GLOBALS['cache_array']['admin'] = $GLOBALS['cache_instance']->getArrayFromCache();

	// Check if valid
	if ((isset($GLOBALS['cache_array']['admin']['login'])) && (is_array($GLOBALS['cache_array']['admin']['login'])) && (is_array($GLOBALS['cache_array']['admin']['admin_id']))) {
		// Check count
		if (count($GLOBALS['cache_array']['admin']['login']) == count($GLOBALS['cache_array']['admin']['admin_id'])) {
			// Rewrite the cache
			$admins = array();
			foreach ($GLOBALS['cache_array']['admin']['login'] as $idx => $admin) {
				// Rewrite all entries
				foreach ($GLOBALS['cache_array']['admin'] as $key => $entry) {
					// Is there login or regular entries?
					if ($key == 'admin_id') {
						// Admin id, so use login
						$admins[$key][$GLOBALS['cache_array']['admin']['login'][$idx]] = $entry[$idx];
					} else {
						// Regular entry so use id
						$admins[$key][$GLOBALS['cache_array']['admin']['admin_id'][$idx]] = $entry[$idx];
					}
				} // END - foreach
			} // END - foreach

			// Transfer back to cache array and remove dummy
			$GLOBALS['cache_array']['admin'] = $admins;
			unset($admins);
		} else {
			// Nope, cache file is corrupted!
			$GLOBALS['cache_instance']->removeCacheFile();
			unset($GLOBALS['cache_array']['admin']);
		}
	} else {
		// Nope, cache file is corrupted!
		$GLOBALS['cache_instance']->removeCacheFile();
		unset($GLOBALS['cache_array']['admin']);
	}
} elseif ((isHtmlOutputMode()) || (isAjaxOutputMode()) || (isRawOutputMode())) {
	// Create cache file
	$GLOBALS['cache_instance']->init();

	// Load every data from DB to cache file
	$add = runFilterChain('sql_admin_extra_data');

	// Query the database about this
	$result = sqlQuery('SELECT
	`id` AS `admin_id`,
	`login`,
	`password`,
	`email`
	' . $add . '
FROM
	`{?_MYSQL_PREFIX?}_admins`
ORDER BY
	`login` ASC', __FILE__, __LINE__);
	while ($row = sqlFetchArray($result)) {
		// Save row
		$GLOBALS['cache_instance']->addRow($row);
	} // END - while

	// Free memory
	sqlFreeResult($result);

	// Close cache
	$GLOBALS['cache_instance']->storeExtensionVersion('admins');
	$GLOBALS['cache_instance']->finalize();
}

// Next cached table are the admins_acls...
if (isExtensionInstalledAndNewer('admins', '0.3')) {
	// Check for cache file
	if (($GLOBALS['cache_instance']->loadCacheFile('admin_acls')) && ($GLOBALS['cache_instance']->extensionVersionMatches('admins'))) {
		// Load admin ACLs from cache
		$GLOBALS['cache_array']['admin_acls'] = $GLOBALS['cache_instance']->getArrayFromCache();

		// Check if valid
		if ((isset($GLOBALS['cache_array']['admin_acls']['admin_id'])) && (is_array($GLOBALS['cache_array']['admin_acls']['admin_id'])) && (isset($GLOBALS['cache_array']['admin_acls']['access_mode'])) && (is_array($GLOBALS['cache_array']['admin_acls']['access_mode']))) {
			// Check count
			if (count($GLOBALS['cache_array']['admin_acls']['admin_id']) == count($GLOBALS['cache_array']['admin_acls']['access_mode'])) {
				// Rewrite the cache
				$admins = array();
				foreach ($GLOBALS['cache_array']['admin_acls']['admin_id'] as $idx => $admin) {
					// Rewrite all entries
					foreach ($GLOBALS['cache_array']['admin_acls'] as $key => $entry) {
						// Is there 'admin_id' or regular entries?
						if ($key != 'admin_id') {
							// Regular entry so use id
							array_push($admins[$key][$GLOBALS['cache_array']['admin_acls']['admin_id'][$idx]], $entry[$idx]);
						} // END - if
					} // END - foreach
				} // END - foreach

				// Transfer back to cache array and remove dummy
				$GLOBALS['cache_array']['admin_acls'] = $admins;
				unset($admins);
			} else {
				// Nope, cache file is corrupted!
				$GLOBALS['cache_instance']->removeCacheFile();
				unset($GLOBALS['cache_array']['admin_acls']);
			}
		} elseif (count($GLOBALS['cache_array']['admin_acls']) > 0) {
			// Nope, cache file is corrupted!
			$GLOBALS['cache_instance']->removeCacheFile();
			unset($GLOBALS['cache_array']['admin_acls']);
		} elseif (isDebugModeEnabled()) {
			// This may drive a lot messages to the logfile
			//* DEBUG: */ logDebugMessage(__FILE__, __LINE__, 'No entry found in admin_acls to rewrite.');
		}
	} elseif ((isHtmlOutputMode()) || (isAjaxOutputMode()) || (isRawOutputMode())) {
		// Create cache file here
		$GLOBALS['cache_instance']->init();

		// Load all admins and their data
		$result = sqlQuery('SELECT * FROM `{?_MYSQL_PREFIX?}_admins_acls` ORDER BY `admin_id` ASC,`action_menu` ASC,`what_menu` ASC', __FILE__, __LINE__);

		// Add all rows
		while ($content = sqlFetchArray($result)) {
			// Add row to cache file
			$GLOBALS['cache_instance']->addRow($content);
		} // END - while

		// Free memory
		sqlFreeResult($result);

		// Close cache
		$GLOBALS['cache_instance']->storeExtensionVersion('admins');
		$GLOBALS['cache_instance']->finalize();
	}
} // END - if

// [EOF]
?>
