<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 07/31/2011 *
 * ===================                          Last change: 07/31/2011 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : load-points_data.php                             *
 * -------------------------------------------------------------------- *
 * Short description : Load more cache files                            *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Mehr Cache-Dateien nachladen                     *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2013 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} elseif (isInstaller()) {
	// Do not run in installation phase
	//* DEBUG: */ logDebugMessage(__FILE__, __LINE__, 'Run from installation phase.');
	return FALSE;
} elseif (!isExtensionInstalledAndNewer('sql_patches', '0.8.0')) {
	// Is not installed or recent enough
	return FALSE;
}

// Let's start with the admins table...
if (($GLOBALS['cache_instance']->loadCacheFile('points_data')) && ($GLOBALS['cache_instance']->extensionVersionMatches('sql_patches'))) {
	// Load cache
	$GLOBALS['cache_array']['points_data'] = $GLOBALS['cache_instance']->getArrayFromCache();

	// Is there cache?
	if (!isset($GLOBALS['cache_array']['points_data']['subject'])) {
		// Try to remove the cache file
		$GLOBALS['cache_instance']->removeCacheFile();

		// Not found, so better abort here
		reportBug(__FILE__, __LINE__, 'points_data has been generated, but does not contain "subject". Please try to reload to fix this.');
	} // END - if

	// Init temporary array
	$pointsData = array();

	// Rewrite all to subject
	foreach ($GLOBALS['cache_array']['points_data']['subject'] as $key => $subject) {
		// Re-add key with subject
		foreach ($GLOBALS['cache_array']['points_data'] as $key2 => $array) {
			// Is key2 not 'subject'?
			if ($key2 != 'subject') {
				// Then Add it
				$pointsData[$subject][$key2] = $GLOBALS['cache_array']['points_data'][$key2][$key];
			} // END - if
		} // END - foreach
	} // END - foreach

	// Set the array back and remove temporary
	$GLOBALS['cache_array']['points_data'] = $pointsData;
	unset($pointsData);
} elseif (((isHtmlOutputMode()) || (isAjaxOutputMode()) || (isRawOutputMode())) && (isExtensionInstalledAndNewer('sql_patches', '0.8.6'))) {
	// Create cache file
	$GLOBALS['cache_instance']->init();

	// Query the database about this
	$result = sqlQuery('SELECT * FROM `{?_MYSQL_PREFIX?}_points_data` ORDER BY `id` ASC', __FILE__, __LINE__);
	while ($data = sqlFetchArray($result)) {
		// Save row
		$GLOBALS['cache_instance']->addRow($data);
	} // END - while

	// Free memory
	sqlFreeResult($result);

	// Close cache
	$GLOBALS['cache_instance']->storeExtensionVersion('sql_patches');
	$GLOBALS['cache_instance']->finalize();
}

// [EOF]
?>
