<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 04/04/2009 *
 * ===================                          Last change: 04/04/2009 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : template-functions.php                           *
 * -------------------------------------------------------------------- *
 * Short description : Template functions                               *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Template-Funktionen                              *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2013 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Wrapper until we merged to the EL branch
function preCompileCode ($code, $full = TRUE) {
	return compileCode($code, $full);
}

// Setter for 'is_template_html'
function enableTemplateHtml ($enable = TRUE) {
	$GLOBALS['is_template_html'] = (bool) $enable;
}

// Checks whether the template is HTML or not by previously set flag
// Default: true
function isTemplateHtml () {
	// Is the output_mode other than 0 (HTML), then no comments are enabled
	if (!isHtmlOutputMode()) {
		// No HTML
		return FALSE;
	} else {
		// Maybe HTML?
		return $GLOBALS['is_template_html'];
	}
}

// Wrapper for writing debug informations to the browser
function debugOutput ($message) {
	outputHtml('<div class="debug_message">' . $message . '</div>');
}

// "Fixes" an empty string into three dashes (use for templates)
function fixEmptyContentToDashes ($str) {
	// Call inner function
	$str = fixNullEmptyToDashes($str, 3);

	// Return string
	return $str;
}

// Init color switch
function initTemplateColorSwitch ($template) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'INIT:' . $template);
	$GLOBALS['color_switch'][$template] = 2;
}

// "Getter" for color switch code
function getColorSwitchCode ($template) {
	// Prepare the code
	$code = "{DQUOTE} . doTemplateColorSwitch('" . $template . "', FALSE, FALSE) . {DQUOTE}";

	// And return it
	return $code;
}

// Output HTML code directly or 'render' it. You addionally switch the new-line character off
function outputHtml ($htmlCode = NULL, $newLine = TRUE) {
	// Init output
	if (!isset($GLOBALS['__output'])) {
		$GLOBALS['__output'] = '';
	} // END - if

	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'getOutputMode()=' . getOutputMode() . ',htmlCode(length)=' . strlen($htmlCode) . ',output(length)=' . strlen($GLOBALS['__output']));
	// Is there HTML-Code here?
	if ((!is_null($htmlCode)) && (!empty($htmlCode))) {
		// Yes, so we handle it as you have configured
		switch (getOutputMode()) {
			case 'render':
				// But if PHP is caching, then we don't need to do that
				if (getPhpCaching() == 'on') {
					// Output into PHP's internal buffer
					outputRawCode($htmlCode);

					// That's why you don't need any \n at the end of your HTML code... :-)
					if ($newLine === TRUE) {
						outputRawCode(PHP_EOL);
					} // END - if
				} else {
					// Render mode for old or lame servers...
					$GLOBALS['__output'] .= $htmlCode;

					// That's why you don't need any \n at the end of your HTML code... :-)
					if ($newLine === TRUE) {
						$GLOBALS['__output'] .= PHP_EOL;
					} // END - if
				}
				break;

			case 'direct':
				// If we are switching from 'render' to 'direct' mode, all data in '__output' must be flushed and cleared
				if ((!empty($GLOBALS['__output'])) && (getPhpCaching() != 'on')) {
					outputRawCode($GLOBALS['__output']);
					$GLOBALS['__output'] = '';
				} // END - if

				// The same as above... ^
				outputRawCode($htmlCode);
				if ($newLine === TRUE) {
					outputRawCode(PHP_EOL);
				} // END - if
				break;

			default:
				// Huh, something goes wrong or maybe you have edited config.php ???
				reportBug(__FUNCTION__, __LINE__, '<strong>{--FATAL_ERROR--}:</strong> {--NO_RENDER_DIRECT--}');
				break;
		} // END - switch
	} elseif ((getPhpCaching() == 'on') && (!isFilledArray($GLOBALS['http_header'])) && (!isRawOutputMode())) {
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'getPhpCaching()=' . getPhpCaching() . ',isset(http_header)=' . intval(isset($GLOBALS['http_header'])) . ',getScriptOutputMode()=' . getScriptOutputMode() . '');
		// Output cached HTML code
		$GLOBALS['__output'] = ob_get_contents();

		// Clear output buffer for later output if output is found
		if (!empty($GLOBALS['__output'])) {
			clearOutputBuffer();
		} // END - if

		// Send all HTTP headers
		sendHttpHeaders();

		// Compile and run finished rendered HTML code
		compileFinalOutput();

		// Output code here, DO NOT REMOVE! ;-)
		outputRawCode($GLOBALS['__output']);
	} elseif ((getOutputMode() == 'render') && (!empty($GLOBALS['__output'])) && (!isRawOutputMode())) {
		// Send all HTTP headers
		sendHttpHeaders();

		// Compile and run finished rendered HTML code
		compileFinalOutput();

		// Output code here, DO NOT REMOVE! ;-)
		outputRawCode($GLOBALS['__output']);
	} else {
		// And flush all headers
		flushHttpHeaders();
	}
}

// Compiles the final output
function compileFinalOutput () {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, '__output(length)=' . strlen($GLOBALS['__output']) . ',getScriptOutputMode()=' . getScriptOutputMode() . ' - ENTERED!');
	// Is this function called?
	if (isset($GLOBALS[__FUNCTION__])) {
		// Abort here
		reportBug(__FUNCTION__, __LINE__, 'Double call of ' . __FUNCTION__ . ' causes problems with sent headers.');
	} // END - if

	// Mark this function as called
	$GLOBALS[__FUNCTION__] = TRUE;

	// Add page header and footer
	addPageHeaderFooter();
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, '__output(length)=' . strlen($GLOBALS['__output']) . ' - After addPageHeaderFooter() call.');

	// Do the final (general) compilation
	$GLOBALS['__output'] = doFinalCompilation($GLOBALS['__output']);

	// Compile any other things out
	$GLOBALS['__output'] = compileUriCode($GLOBALS['__output']);

	// Extension 'rewrite' installed?
	if ((isExtensionActive('rewrite')) && (!isCssOutputMode())) {
		$GLOBALS['__output'] = rewriteLinksInCode($GLOBALS['__output']);
	} // END - if

	// Compress it?
	/**
	 * @TODO On some pages this is buggy
	if (!empty($_SERVER['HTTP_ACCEPT_ENCODING']) && (isInStringIgnoreCase('gzip', $_SERVER['HTTP_ACCEPT_ENCODING']))) {
		// Compress it for HTTP gzip
		$GLOBALS['__output'] = gzencode($GLOBALS['__output'], 9);

		// Add header
		addHttpHeader('Content-Encoding: gzip');
	} elseif (!empty($_SERVER['HTTP_ACCEPT_ENCODING']) && (isInStringIgnoreCase('deflate', $_SERVER['HTTP_ACCEPT_ENCODING']))) {
		// Compress it for HTTP deflate
		$GLOBALS['__output'] = gzcompress($GLOBALS['__output'], 9);

		// Add header
		addHttpHeader('Content-Encoding: deflate');
	}
	*/

	// Add final length
	addHttpHeader('Content-Length: ' . strlen($GLOBALS['__output']));

	// Flush all headers
	flushHttpHeaders();
}

// Main compilation loop
function doFinalCompilation ($code, $insertComments = TRUE, $enableCodes = TRUE) {
	// Code must not be an array (happens in installer)
	assert(!is_array($code));

	// Insert comments? (Only valid with HTML templates, of course)
	enableTemplateHtml($insertComments);

	// Init counter
	$totalCompilations = 0;

	// Compile all out
	while (((isInString('{--', $code)) || (isInString('{DQUOTE}', $code)) || (isInString('{?', $code)) || (isInString('{%', $code) !== FALSE)) && ($totalCompilations < 7)) {
		// Init common variables
		$content = array();
		$newContent = '';

		// Compile it
		//* DEBUG: */ debugOutput('<pre>'.lineNumberCode($code).'</pre>');
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code[]=' . gettype($code) . ',enableCodes[]=' . gettype($enableCodes));
		$eval = '$newContent = "' . str_replace('{DQUOTE}', '"', compileCode(escapeQuotes($code), $enableCodes)) . '";';
		//* DEBUG: */ if (!$insertComments) print('EVAL=<pre>'.lineNumberCode($eval).'</pre>');
		eval($eval);
		//* DEBUG: */ if (!$insertComments) print('NEW=<pre>'.lineNumberCode($newContent).'</pre>');
		//* DEBUG: */ die('<pre>'.encodeEntities($newContent).'</pre>');

		// Was that eval okay?
		if (empty($newContent)) {
			// Something went wrong!
			reportBug(__FUNCTION__, __LINE__, 'Evaluation error:<pre>' . lineNumberCode($eval) . '</pre>', FALSE);
		} // END - if

		// Use it again
		$code = $newContent;

		// Compile the final code if insertComments is true
		if ($insertComments == TRUE) {
			// ... because SQL queries shall keep OPEN_CONFIG and such in
			$code = compileRawCode($code);
		} // END - if

		// Count round
		$totalCompilations++;
	} // END - while

	// Add debugging data in HTML code, if mode is enabled
	if ((isDebugModeEnabled()) && ($insertComments === TRUE) && (isHtmlOutputMode())) {
		// Add loop count
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'isDebugModeEnabled()=' . intval(isDebugModeEnabled()) . ',insertComments=' . intval($insertComments) . ',isHtmlOutputMode()=' . intval(isHtmlOutputMode()));
		$code .= '<!-- Total compilation loop=' . $totalCompilations . ' //-->';
	} // END - if

	// Return the compiled code
	return $code;
}

// Output the raw HTML code
function outputRawCode ($htmlCode) {
	// Output stripped HTML code to avoid broken JavaScript code, etc.
	print(str_replace('{BACK}', chr(92), $htmlCode));

	// Flush the output if only getPhpCaching() is not 'on'
	if (getPhpCaching() != 'on') {
		// Flush it
		flush();
	} // END - if
}

// Load a template file and return it's content (only it's name; do not use ' or ")
function loadTemplate ($template, $return = FALSE, $content = array(), $compileCode = TRUE) {
	// @TODO Remove these sanity checks if all is fine
	if (!is_bool($return)) {
		// $return has to be boolean
		reportBug(__FUNCTION__, __LINE__, 'return[] is not bool (' . gettype($return) . ')');
	} elseif (!is_string($template)) {
		// $template has to be string
		reportBug(__FUNCTION__, __LINE__, 'template[] is not string (' . gettype($template) . ')');
	}

	// Init returned content
	$templateContent = '';

	// Set current template
	$GLOBALS['current_template'] = $template;

	// Is there cache?
	if ((!isDebugTemplateCacheEnabled()) && (isTemplateCached('html', $template))) {
		// Evaluate the cache
		$templateContent = readTemplateCache('html', $template, $content);

		// Better remove array element which is only needed in uncached mode
		unset($GLOBALS['template_eval']['html'][$template]);
	} elseif (!isset($GLOBALS['template_eval']['html'][$template])) {
		// Make all template names lowercase
		$template = strtolower($template);

		// Base directory
		$basePath = getTemplateBasePath('html');
		$extraPath = detectExtraTemplatePath('html', $template);

		// Generate FQFN
		$FQFN = $basePath . '/' . $extraPath . $template . '.tpl';
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Template ' . $template . ' is solved to FQFN=' . $FQFN);

		// Does the special template exists?
		if (!isFileReadable($FQFN)) {
			// Reset to default template
			$FQFN = $basePath . '/' . $template . '.tpl';
		} // END - if

		// Now does the final template exists?
		if (isFileReadable($FQFN)) {
			// Count the template load
			incrementConfigEntry('num_templates');

			// The local file does exists so we load it. :)
			$GLOBALS['template_content']['html'][$template] = readFromFile($FQFN);

			// Is there to compile the code?
			if ((isInString('$', $GLOBALS['template_content']['html'][$template])) || (isInString('{--', $GLOBALS['template_content']['html'][$template])) || (isInString('{?', $GLOBALS['template_content']['html'][$template])) || (isInString('{%', $GLOBALS['template_content']['html'][$template]))) {
				// Normal HTML output?
				if ((isHtmlOutputMode()) && (substr($template, 0, 3) != 'js_')) {
					// Add surrounding HTML comments to help finding bugs faster
					$code = '<!-- Template ' . $template . ' - Start //-->' . $GLOBALS['template_content']['html'][$template] . '<!-- Template ' . $template . ' - End //-->';

					// Prepare eval() command
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
					$GLOBALS['template_eval']['html'][$template] = '$templateContent = "' . getColorSwitchCode($template) . compileCode(escapeQuotes($code), TRUE, $compileCode) . '";';
				} elseif (substr($template, 0, 3) == 'js_') {
					// JavaScripts don't like entities, dollar signs and timings
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
					$GLOBALS['template_eval']['html'][$template] = '$templateContent = decodeEntities("' . escapeJavaScriptQuotes($GLOBALS['template_content']['html'][$template]) . '");';
				} elseif (isAjaxOutputMode()) {
					// AJAX (JSON content)
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
					$GLOBALS['template_eval']['html'][$template] = '$templateContent = "' . escapeJavaScriptQuotes($GLOBALS['template_content']['html'][$template]) . '";';
				} else {
					// Prepare eval() command, other output doesn't like entities, maybe
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
					$GLOBALS['template_eval']['html'][$template] = '$templateContent = decodeEntities("' . compileRawCode(escapeQuotes($GLOBALS['template_content']['html'][$template]), TRUE, $compileCode) . '");';
				}
			} elseif (isHtmlOutputMode()) {
				// Add surrounding HTML comments to help finding bugs faster
				$templateContent = '<!-- Template ' . $template . ' - Start //-->' . $GLOBALS['template_content']['html'][$template] . '<!-- Template ' . $template . ' - End //-->';
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
				$GLOBALS['template_eval']['html'][$template] = '$templateContent = "' . getColorSwitchCode($template) . compileRawCode(escapeQuotes($templateContent), TRUE, $compileCode) . '";';
			} elseif (isAjaxOutputMode()) {
				// AJAX (JSON content)
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
				$GLOBALS['template_eval']['html'][$template] = '$templateContent = "' . escapeJavaScriptQuotes($GLOBALS['template_content']['html'][$template]) . '";';
			} else {
				// JavaScript again
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
				$GLOBALS['template_eval']['html'][$template] = '$templateContent = decodeEntities("' . escapeJavaScriptQuotes($GLOBALS['template_content']['html'][$template]) . '");';
			} // END - if
		} elseif ((isAdmin()) || ((isInstalling()) && (!isInstalled()))) {
			// Only admins shall see this warning or when installation mode is active
			$templateContent = '<div class="para">
	{--TEMPLATE_404--}
</div>
<div class="para">
	(' . $template . ')
</div>
<div class="para">
	{--TEMPLATE_CONTENT--}:
	<pre>' . print_r($content, TRUE) . '</pre>
</div>';
		} else {
			// No file!
			$GLOBALS['template_eval']['html'][$template] = '404';
		}
	}

	// Code set?
	if ((isset($GLOBALS['template_eval']['html'][$template])) && ($GLOBALS['template_eval']['html'][$template] != '404')) {
		// Eval the code
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'template=' . $template . ' - BEFORE EVAL');
		///* DEBUG: */ print('<pre>'.htmlentities($GLOBALS['template_eval']['html'][$template]).'</pre>');
		eval($GLOBALS['template_eval']['html'][$template]);
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'template=' . $template . ' - AFTER EVAL');
	} // END - if

	// Is there some content to output or return?
	if ((empty($templateContent)) && (isDebugModeEnabled())) {
		// Warning, empty output!
		return 'E:' . $template . ',content=<pre>' . print_r($content, TRUE) . '</pre>';
	} // END - if

	// Not empty so let's put it out! ;)
	if ($return === TRUE) {
		// Return the HTML code
		return $templateContent;
	} else {
		// Output directly
		outputHtml($templateContent);
	}
}

// Detects the extra template path from given template name
function detectExtraTemplatePath ($prefix, $template) {
	// Default is empty
	$extraPath = '';

	// Is there cache?
	if (!isset($GLOBALS['extra_path'][$prefix][$template])) {
		// Check for admin/guest/member/etc. templates
		if (substr($template, 0, 6) == 'admin_') {
			// Admin template found
			$extraPath = 'admin/';
		} elseif (substr($template, 0, 6) == 'guest_') {
			// Guest template found
			$extraPath = 'guest/';
		} elseif (substr($template, 0, 7) == 'member_') {
			// Member template found
			$extraPath = 'member/';
		} elseif (substr($template, 0, 7) == 'select_') {
			// Selection template found
			$extraPath = 'select/';
		} elseif (substr($template, 0, 8) == 'install_') {
			// Installation template found
			$extraPath = 'install/';
		} elseif (substr($template, 0, 4) == 'ext_') {
			// Extension template found
			$extraPath = 'ext/';
		} elseif (substr($template, 0, 3) == 'la_') {
			// 'Logical-area' template found
			$extraPath = 'la/';
		} elseif (substr($template, 0, 3) == 'js_') {
			// JavaScript template found
			$extraPath = 'js/';
		} elseif (substr($template, 0, 5) == 'menu_') {
			// Menu template found
			$extraPath = 'menu/';
		} else {
			// Test for extension
			$test = substr($template, 0, strpos($template, '_'));

			// Probe for valid extension name
			if (isExtensionNameValid($test)) {
				// Set extra path to extension's name
				$extraPath = $test . '/';
			} // END - if
		}

		// Store it in cache
		$GLOBALS['extra_path'][$prefix][$template] = $extraPath;
	} // END - if

	// Return result
	return $GLOBALS['extra_path'][$prefix][$template];
}

// Loads an email template and compiles it
function loadEmailTemplate ($template, $content = array(), $userid = NULL, $loadUserData = TRUE) {
	// Make sure all template names are lowercase!
	$template = strtolower($template);

	// Set current template
	$GLOBALS['current_template'] = $template;

	// Is content an array?
	if (is_array($content)) {
		// Add expiration to array
		if ((isExtensionInstalled('autopurge')) && (isConfigEntrySet('auto_purge')) && (getAutoPurge() == '0')) {
			// Will never expire!
			$content['expiration'] = '{--MAIL_WILL_NEVER_EXPIRE--}';
		} elseif ((isExtensionInstalled('autopurge')) && (isConfigEntrySet('auto_purge'))) {
			// Create nice date string
			$content['expiration'] = '{%config,createFancyTime=auto_purge%}';
		} else {
			// Missing entry
			$content['expiration'] = '{--MAIL_NO_CONFIG_AUTO_PURGE--}';
		}
	} // END - if

	// Is there cache?
	if ((!isDebugTemplateCacheEnabled()) && (isTemplateCached('email', $template))) {
		// Evaluate the cache
		$templateContent = readTemplateCache('email', $template, $content);

		// Better remove array element which is need only in uncached mode
		unset($GLOBALS['template_eval']['email'][$template]);
	} elseif (!isset($GLOBALS['template_eval']['email'][$template])) {
		// Base directory
		$basePath = getTemplateBasePath('emails');

		// Detect extra path
		$extraPath = detectExtraTemplatePath('email', $template);

		// Generate full FQFN
		$FQFN = $basePath . '/' . $extraPath . $template . '.tpl';

		// Does the special template exists?
		if (!isFileReadable($FQFN)) {
			// Reset to default template
			$FQFN = $basePath . '/' . $template . '.tpl';
		} // END - if

		// Now does the final template exists?
		$templateContent = '';
		if (isFileReadable($FQFN)) {
			// The local file does exists so we load it. :)
			$GLOBALS['template_content']['email'][$template] = readFromFile($FQFN);

			// Run code
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Reached!');
			$GLOBALS['template_eval']['email'][$template] = '$templateContent = decodeEntities("' . compileRawCode(escapeQuotes($GLOBALS['template_content']['email'][$template])) . '");';
		} elseif (!empty($template)) {
			// Template file not found
			$templateContent = '<div class="para">
	{--TEMPLATE_404--}: ' . $template . '
</div>
<div class="para">
	{--TEMPLATE_CONTENT--}:
	<pre>' . print_r($content, TRUE) . '</pre>
</div>';

			// Don't cache this, as there is no template to cache
			$GLOBALS['template_eval']['email'][$template] = '404';

			// Debug mode not active? Then remove the HTML tags
			if (!isDebugModeEnabled()) {
				// Remove HTML tags
				$templateContent = secureString($templateContent);
			} // END - if
		} else {
			// No template name supplied!
			$templateContent = '{--NO_TEMPLATE_SUPPLIED--}';
			$GLOBALS['template_eval']['email'][$template] = '404';
		}
	}

	// Is there something to eval?
	if ((isset($GLOBALS['template_eval']['email'][$template])) && ($GLOBALS['template_eval']['email'][$template] != '404')) {
		// Eval the code
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'template=' . $template . ' - BEFORE EVAL');
		//* DEBUG: */ print('<pre>'.htmlentities($GLOBALS['template_eval']['email'][$template]).'</pre>');
		//* DEBUG: */ die('<pre>'.print_r($content, TRUE).'</pre><pre>'.htmlentities($GLOBALS['template_eval']['email'][$template]).'</pre>');
		eval($GLOBALS['template_eval']['email'][$template]);
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'template=' . $template . ' - AFTER EVAL');
	} // END - if

	// Are there some content?
	if (empty($templateContent)) {
		// Compiling failed
		$templateContent = "Compiler error for template " . $template . " !\nUncompiled content:\n" . $GLOBALS['template_eval']['email'][$template];

		// Add last error if the required function exists
		if (function_exists('error_get_last')) {
			// Add last error and some lines for better overview
			$templateContent .= "\n--------------------------------------\nDebug:\n" . print_r(error_get_last(), TRUE) . "--------------------------------------\nPlease don't alter these informations!\nThanx.";
		} // END - if
	} // END - if

	// Remove content and data
	unset($content);

	// Return content
	return $templateContent;
}

// "Getter" for menu CSS classes, mainly used in templates
function getMenuCssClasses ($data) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'data=' . $data);

	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$data])) {
		// $data needs to be converted into an array
		$content = explode('|', $data);

		// Non-existent index 2 will happen in menu blocks
		if (!isset($content[2])) {
			$content[2] = '';
		} // END - if

		// Re-construct the array: 0=visible,1=locked,2=prefix
		$content['visible'] = $content[0];
		$content['locked']  = $content[1];

		// Call our "translator" function
		$content = translateMenuVisibleLocked($content, $content[2]);

		// Set it in cache
		$GLOBALS[__FUNCTION__][$data] = ($content['visible_css'] . ' ' . $content['locked_css']);
	} // END - if

	// Return cache
	return $GLOBALS[__FUNCTION__][$data];
}

// Generate XHTML code for the CAPTCHA
function generateCaptchaCode ($code, $type, $urlId, $userid) {
	return '<img border="0" alt="Code ' . $code . '" src="{%url=mailid.php?userid=' . $userid . '&amp;' . $type . '=' . $urlId . '&amp;do=img&amp;code=' . $code . '%}" />';
}

// Compiles the given HTML/mail code
function compileCode ($code, $full = TRUE, $compileCode = TRUE) {
	// Is the code a string or should we not compile?
	if ((!is_string($code)) || ($compileCode === FALSE)) {
		// Silently return it
		return $code;
	} // END - if

	// Start couting
	$startCompile = microtime(TRUE);

	// Comile the code
	$code = compileRawCode($code, $full, $compileCode);

	// Get timing
	$compilationTime = $startCompile - microtime(TRUE);

	// Add timing if enabled
	if (isTemplateHtml()) {
		// Add timing, this should be disabled in
		$code .= '<!-- Compilation time: ' . ($compilationTime * 1000). 'ms //-->';
	} // END - if

	// Return compiled code
	return $code;
}

// Compiles the code
function compileRawCode ($code, $full = TRUE, $compileCode = TRUE) {
	//* DIE: */ reportBug(__FUNCTION__, __LINE__, 'Called with ' . strlen($code) . ' code length.');
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Called with code()=' . strlen($code) . ',full=' . intval($full) . ',compileCode=' . intval($compileCode));

	// Is the code a string or shall we not compile?
	if ((!is_string($code)) || ($compileCode === FALSE)) {
		// Silently return it
		//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code[]= ' . gettype($code) . ' is not a string or compileCode(' . intval($compileCode) . ') is FALSE.');
		return $code;
	} // END - if

	// Init replacement-array with smaller set of security characters
	$secChars = $GLOBALS['url_chars'];

	// Select full set of chars to replace when we e.g. want to compile URLs
	if ($full === TRUE) {
		$secChars = $GLOBALS['security_chars'];
	} // END - if

	// First compile these chars
	array_unshift($secChars['to']  , '{--'       , '--}');
	array_unshift($secChars['from'], '{%message,', '%}' );

	// Replace QUOT and other non-HTML codes
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code()=' . strlen($code) . ' - before str_replace() ...');
	$code = str_replace($secChars['to'], $secChars['from'], $code);
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code()=' . strlen($code) . ' - after str_replace() ...');

	// Compile the prepared code through a filter chain
	$code = runFilterChain('compile_code', $code);

	// Find all $content[bla][blub] entries
	preg_match_all('/\$content((\[([a-zA-Z0-9-_]+)\])*)/', $code, $matches);
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Second regex gave ' . count($matches[0]) . ' matches.');

	// Are some matches found?
	if ((isFilledArray($matches)) && (isFilledArray($matches[0]))) {
		// Replace all matches
		$matchesFound = array();
		foreach ($matches[0] as $key => $match) {
			// Fuzzy look has failed by default
			$fuzzyFound = FALSE;

			// "Cache" match length
			$matchLength = strlen($match);

			// Fuzzy look on match if already found
			foreach ($matchesFound as $found => $set) {
				// Get test part
				$test = substr($found, 0, $matchLength);

				// Does this entry exist?
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'found=' . $found . ',match=' . $match . ',set=' . $set);
				if ($test == $match) {
					// Match found
					//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'fuzzyFound!');
					$fuzzyFound = TRUE;
					break;
				} // END - if
			} // END - foreach

			// Skip this entry?
			if ($fuzzyFound === TRUE) {
				continue;
			} // END - if

			// Take all string elements
			if ((is_string($matches[3][$key])) && (!isset($matchesFound[$match])) && (!isset($matchesFound[$key.'_' . $matches[3][$key]]))) {
				// Replace it in the code, replace dollar sign so it won't be detected by next regex (see there)
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',match=' . $match);
				$newMatch = str_replace(array('[', ']', '$'), array("['", "']", '{COMPILE_DOLLAR}'), $match);
				$code = str_replace($match, '".' . $newMatch . '."', $code);
				$matchesFound[$key . '_' . $matches[3][$key]] = 1;
				$matchesFound[$match] = TRUE;
			} elseif (!isset($matchesFound[$match])) {
				// Not yet replaced!
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'match=' . $match);
				$code = str_replace($match, '".' . $match . '."', $code);
				$matchesFound[$match] = 1;
			} else {
				// Everthing else should be a least logged
				logDebugMessage(__FUNCTION__, __LINE__, 'match=' . $match . ',key=' . $key);
			}
		} // END - foreach
	} // END - if

	/*
	 * Find $foobar, $foo_bar and $fooBar entries. This regex would also find
	 * $content[foo_bar] which would result in {DOLLAR}content[foo_bar] and
	 * therefore the variable's value won't be inserted. This is why
	 * {COMPILE_DOLLAR} is being used in above loop and at the end of this
	 * function being replace with the original dollar sign again.
	 */
	preg_match_all('/\$([a-z_A-Z\[\]]){0,}/', $code, $matches);

	// Are some matches found?
	if ((isFilledArray($matches)) && (isFilledArray($matches[0]))) {
		// Scan all matches for not $content
		foreach ($matches[0] as $match) {
			// Trim match
			$match = trim($match);

			// Debug message
			//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'match=' . $match);

			// Is the first part not $content/$userid and not empty?
			// @TODO $userid is deprecated and should be removed from loadEmailTemplate() and replaced with $content[userid] in all templates
			if ((!empty($match)) && (substr($match, 0, 8) != '$content') && ($match != '$userid')) {
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'match=' . $match . ' - SECURED!');
				// Then replace $ with {DOLLAR}
				$matchSecured = str_replace('$', '{DOLLAR}', $match);

				// And in $code as well
				$code = str_replace($match, $matchSecured, $code);
			} // END - if
		} // END - if
	} // END - if

	// Add 'COMPILE_DOLLAR' again
	array_push($secChars['to']  , '{COMPILE_DOLLAR}');
	array_push($secChars['from'], '$');

	// Replace QUOT and other non-HTML codes
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code()=' . strlen($code) . ' - before str_replace() ...');
	$code = str_replace($secChars['to'], $secChars['from'], $code);
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code()=' . strlen($code) . ' - after str_replace() ...');

	// Finally return it
	//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Exiting with code()=' . strlen($code));
	return $code;
}

//
function addSelectionBox ($type, $default, $prefix = '', $id = NULL, $class = 'form_select') {
	$OUT = '';

	if ($type == 'yn') {
		// This is a yes/no selection only!
		if (isValidId($id)) $prefix .= '[' . $id . ']';
		$OUT .= '<select name="' . $prefix . '" class="' . $class . '" size="1">';
	} else {
		// Begin with regular selection box here
		if (!empty($prefix)) $prefix .= '_';
		$type2 = $type;
		if (isValidId($id)) $type2 .= '[' . $id . ']';
		$OUT .= '<select name="' . strtolower($prefix . $type2) . '" class="' . $class . '" size="1">';
	}

	switch ($type) {
		case 'ye': // Years
			// Get current year
			$year = getYear();

			// Use configured min age or fixed?
			if (isExtensionInstalledAndNewer('other', '0.2.1')) {
				// Configured
				$startYear = $year - getMinAge();
			} else {
				// Fixed 16 years
				$startYear = $year - 16;
			}

			// Calculate earliest year (100 years old people can still enter Internet???)
			$minYear = $year - 100;

			// Check if the default value is larger than minimum and bigger than actual year
			if (($default > $minYear) && ($default >= $year)) {
				for ($idx = $year; $idx < ($year + 11); $idx++) {
					$OUT .= '<option value="' . $idx . '"';
					if ($default == $idx) $OUT .= ' selected="selected"';
					$OUT .= '>' . $idx . '</option>';
				} // END - for
			} elseif ($default == -1) {
				// Current year minus 1
				for ($idx = $startYear; $idx <= ($year + 1); $idx++) {
					$OUT .= '<option value="' . $idx . '">' . $idx . '</option>';
				} // END - for
			} else {
				// Get current year and subtract the configured minimum age
				$OUT .= '<option value="' . ($minYear - 1) . '">&lt;' . $minYear . '</option>';

				// Construct year selection list
				for ($idx = $minYear; $idx <= $startYear; $idx++) {
					$OUT .= '<option value="' . $idx . '"';
					if ($default == $idx) $OUT .= ' selected="selected"';
					$OUT .= '>' . $idx . '</option>';
				} // END - for
			}
			break;

		case 'mo': // Months
			foreach ($GLOBALS['month_descr'] as $idx => $descr) {
				$OUT .= '<option value="' . $idx . '"';
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= '>' . $descr . '</option>';
			} // END - for
			break;

		case 'mn': // Months, numeric
			for ($idx = 0; $idx <= 12; $idx++) {
				$OUT .= '<option value="' . $idx . '"';
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= '>{%pipe,padLeftZero=' . $idx . '%}</option>';
			} // END - for
			break;

		case 'we': // Weeks
			for ($idx = 0; $idx <= 4; $idx++) {
				$OUT .= '  <option value="' . $idx . '"';
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= '>' . $idx . '</option>';
			} // END - for
			break;

		case 'da': // Days
			for ($idx = 0; $idx <= 31; $idx++) {
				$OUT .= '<option value="' . $idx . '"';
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= '>{%pipe,padLeftZero=' . $idx . '%}</option>';
			} // END - for
			break;

		case 'ho': // Hours
			for ($idx = 0; $idx <= 23; $idx++) {
				$padded = padLeftZero($idx, 2);
				$OUT .= '<option value="' . $padded . '"';
				if ($default == $padded) $OUT .= ' selected="selected"';
				$OUT .= '>' . $padded . '</option>';
			} // END - for
			break;

		case 'mi': // Minutes
		case 'se': // Seconds
			for ($idx = 0; $idx <= 59; $idx+=5) {
				$padded = padLeftZero($idx, 2);
				$OUT .= '<option value="' . $padded . '"';
				if ($default == $padded) $OUT .= ' selected="selected"';
				$OUT .= '>' . $padded . '</option>';
			} // END - for
			break;

		case 'yn': // Yes/No
			$OUT .= '<option value="Y"';
			if ($default == 'Y') $OUT .= ' selected="selected"';
			$OUT .= '>{--YES--}</option><option value="N"';
			if ($default != 'Y') $OUT .= ' selected="selected"';
			$OUT .= '>{--NO--}</option>';
			break;

		default: // Not detected
			reportBug(__FUNCTION__, __LINE__, 'type=' . $type . ',default=' . $default . ',prefix=' .  $prefix . ',id[' . gettype($id) . ']=' . $id . ',class=' . $class . ' - is not supported.');
			break;
	}
	$OUT .= '</select>';
	return $OUT;
}

// Insert the code in $img_code into jpeg or PNG image
function generateImageOrCode ($img_code, $headerSent = TRUE) {
	// Is the code size oversized or shouldn't we display it?
	if ((strlen($img_code) > 6) || (empty($img_code)) || (getCodeLength() == '0')) {
		// Stop execution of function here because of over-sized code length
		reportBug(__FUNCTION__, __LINE__, 'img_code ' . $img_code .' has invalid length. img_code(length)=' . strlen($img_code) . ' code_length=' . getCodeLength());
	} elseif ($headerSent === FALSE) {
		// Return an HTML code here
		return '<img src="{%url=img.php?code=' . $img_code . '%}" alt="Image" />';
	}

	// Load image
	$img = sprintf('%s/theme/%s/images/code_bg.%s',
		getPath(),
		getCurrentTheme(),
		getImgType()
	);

	// Is it readable?
	if (isFileReadable($img)) {
		// Switch image type
		switch (getImgType()) {
			case 'jpg': // Okay, load image and hide all errors
				$image = imagecreatefromjpeg($img);
				break;

			case 'png': // Okay, load image and hide all errors
				$image = imagecreatefrompng($img);
				break;
		} // END - switch
	} else {
		// Silently log the error
		logDebugMessage(__FUNCTION__, __LINE__, sprintf('File for image-type %s in theme %s not found.', getImgType(), getCurrentTheme()));
		return;
	}

	// Generate text color (red/green/blue; 0 = dark, 255 = bright)
	$text_color = imagecolorallocate($image, 0, 0, 0);

	// Insert code into image
	imagestring($image, 5, 14, 2, $img_code, $text_color);

	// Return to browser
	setContentType('image/' . getImgType());

	// Output image with matching image factory
	switch (getImgType()) {
		case 'jpg': imagejpeg($image); break;
		case 'png': imagepng($image);  break;
	} // END - switch

	// Remove image from memory
	imagedestroy($image);
}

// Create selection box or array of splitted timestamp
function createTimeSelections ($timestamp, $prefix = '', $display = '', $align = 'center', $asArray = FALSE) {
	// Do not continue if ONE_DAY is absend
	if (!isConfigEntrySet('ONE_DAY')) {
		// Abort here
		reportBug(__FUNCTION__, __LINE__, 'Configuration entry ONE_DAY is absend. timestamp=' . $timestamp . ',prefix=' . $prefix . ',align=' . $align . ',asArray=' . intval($asArray));
	} // END - if

	// Calculate 2-seconds timestamp
	$stamp = round($timestamp);
	//* DEBUG: */ debugOutput('*' . $stamp .'/' . $timestamp . '*');

	// Is there a leap year?
	$SWITCH = '0';
	$TEST = getYear() / 4;
	$M1 = getMonth();
	$M2 = getMonth(time() + $timestamp);

	// If so and if current time is before 02/29 and estimated time is after 02/29 then add 86400 seconds (one day)
	if ((floor($TEST) == $TEST) && ($M1 == '02') && ($M2 > '02'))  {
		$SWITCH = getOneDay();
	} // END - switch

	// First of all years...
	$year = abs(floor($timestamp / (31536000 + $SWITCH)));
	//* DEBUG: */ debugOutput('year=' . $year);
	// Next months...
	$month = abs(floor($timestamp / 2628000 - $year * 12));
	//* DEBUG: */ debugOutput('month=' . $month);
	// Next weeks
	$week = abs(floor($timestamp / 604800 - $year * ((365 + $SWITCH / getOneDay()) / 7) - ($month / 12 * (365 + $SWITCH / getOneDay()) / 7)));
	//* DEBUG: */ debugOutput('week=' . $week);
	// Next days...
	$day = abs(floor($timestamp / 86400 - $year * (365 + $SWITCH / getOneDay()) - ($month / 12 * (365 + $SWITCH / getOneDay())) - $week * 7));
	//* DEBUG: */ debugOutput('day=' . $day);
	// Next hours...
	$hour = abs(floor($timestamp / 3600 - $year * (365 + $SWITCH / getOneDay()) * 24 - ($month / 12 * (365 + $SWITCH / getOneDay()) * 24) - $week * 7 * 24 - $day * 24));
	//* DEBUG: */ debugOutput('hour=' . $hour);
	// Next minutes..
	$minute = abs(floor($timestamp / 60 - $year * (365 + $SWITCH / getOneDay()) * 24 * 60 - ($month / 12 * (365 + $SWITCH / getOneDay()) * 24 * 60) - $week * 7 * 24 * 60 - $day * 24 * 60 - $hour * 60));
	//* DEBUG: */ debugOutput('minute=' . $minute);
	// And at last seconds...
	$second = abs(floor($timestamp - $year * (365 + $SWITCH / getOneDay()) * 24 * 3600 - ($month / 12 * (365 + $SWITCH / getOneDay()) * 24 * 3600) - $week * 7 * 24 * 3600 - $day * 24 * 3600 - $hour * 3600 - $minute * 60));
	//* DEBUG: */ debugOutput('second=' . $second);

	// Is seconds zero and time is < 60 seconds?
	if (($second < 1) && ($timestamp < 60)) {
		// Fix seconds
		$second = round($timestamp);
	} // END - if

	// Put all calculated values in array
	$data = array(
		'Y' => $year,
		'M' => $month,
		'W' => $week,
		'D' => $day,
		'h' => $hour,
		'm' => $minute,
		's' => $second
	);

	//
	// Now we convert them in seconds...
	//
	if ($asArray === TRUE) {
		// Just put data array out
		$OUT = $data;
	} else {
		// Init array
		$units = array(
			// Time unit -> field name
			'unit_field' => array(
				'Y' => 'ye',
				'M' => 'mn',
				'W' => 'we',
				'D' => 'da',
				'h' => 'ho',
				'm' => 'mi',
				's' => 'se'
			),
			// Time unit -> label
			'unit_label' => array(
				'Y' => 'YEAR',
				'M' => 'MONTH',
				'W' => 'WEEK',
				'D' => 'DAY',
				'h' => 'HOUR',
				'm' => 'MINUTE',
				's' => 'SECOND'
			)
		);

		// Generate table
		$OUT  = '<div align="' . $align . '">';
		$OUT .= '<table border="0" cellspacing="0" cellpadding="0" class="timebox_table dashed">';
		$OUT .= '<tr>';

		// "Walk" through all units
		foreach ($units['unit_field'] as $unit => $field) {
			// Is this displayed or zero?
			if (isInString($unit, $display) || (empty($display))) {
				// @TODO <label for="' . $prefix . '_' . $field . '"></<label> not working here
				$OUT .= '<td align="center" class="timebox_column bottom"><div class="tiny">{--TIME_UNIT_' . $units['unit_label'][$unit] . '--}</div></td>';
			} // END - if
		} // END - foreach

		// Close table row and open new one
		$OUT .= '</tr>';
		$OUT .= '<tr>';

		// "Walk" through all units again
		foreach ($units['unit_field'] as $unit => $field) {
			// Is this used?
			if (isInString($unit, $display) || (empty($display))) {
				// Generate year selection
				$OUT .= '<td align="center">';
				$OUT .= addSelectionBox($field, $data[$unit], $prefix, NULL, 'mini_select');
				$OUT .= '</td>';
			} else {
				$OUT .= '<input type="hidden" name="' . $prefix . '_' . $field . '" value="0" />';
			}
		} // END - foreach

		// Close all tags
		$OUT .= '</tr>';
		$OUT .= '</table>';
		$OUT .= '</div>';
	}

	// Return generated HTML code or data array
	return $OUT;
}

// Generate a list of administrative links to a given userid
function generateMemberAdminActionLinks ($userid) {
	// Make sure userid is a number
	if ($userid != bigintval($userid)) {
		reportBug(__FUNCTION__, __LINE__, 'userid is not a number!');
	} // END - if

	// Define all main targets
	$targetArray = runFilterChain('member_admin_actions', array('del_user', 'edit_user', 'lock_user', 'list_refs', 'list_links', 'add_points', 'sub_points'));

	// Get user status
	$status = getFetchedUserData('userid', $userid, 'status');

	// Begin of navigation links
	$OUT = '[';

	foreach ($targetArray as $target) {
		$OUT .= '<span class="admin_user_link"><a href="{%url=modules.php?module=admin&amp;what=' . $target . '&amp;userid=' . $userid . '%}" title="{--ADMIN_USER_ACTION_LINK_';
		//* DEBUG: */ debugOutput('*' . $target.'/' . $status.'*');
		if (($target == 'lock_user') && ($status == 'LOCKED')) {
			// Locked accounts shall be unlocked
			$OUT .= 'UNLOCK_USER';
		} elseif ($target == 'del_user') {
			// @TODO Deprecate this thing
			$OUT .= 'DELETE_USER';
		} else {
			// All other status is fine
			$OUT .= strtoupper($target);
		}
		$OUT .= '_TITLE--}">{--ADMIN_USER_ACTION_LINK_';
		if (($target == 'lock_user') && ($status == 'LOCKED')) {
			// Locked accounts shall be unlocked
			$OUT .= 'UNLOCK_USER';
		} elseif ($target == 'del_user') {
			// @TODO Deprecate this thing
			$OUT .= 'DELETE_USER';
		} else {
			// All other status is fine
			$OUT .= strtoupper($target);
		}
		$OUT .= '--}</a></span>|';
	} // END - foreach

	// Add special link, in case of the account is unconfirmed
	if ($status == 'UNCONFIRMED') {
		// Add it
		$OUT .= '<span class="admin_user_link"><a target="_blank" title="{--ADMIN_USER_ACTION_LINK_CONFIRM_ACCOUNT_TITLE--}" href="{%url=confirm.php?hash=' . getFetchedUserData('userid', $userid, 'user_hash') . '%}">{--ADMIN_USER_ACTION_LINK_CONFIRM_ACCOUNT--}</a></span>|';
	} // END - if

	// Finish navigation link
	$OUT = substr($OUT, 0, -1) . ']';

	// Return string
	return $OUT;
}

// Generate an email link
function generateEmailLink ($email, $table = 'admins') {
	// Default email link (INSECURE! Spammer can read this by harvester programs)
	$EMAIL = 'mailto:' . $email;

	// Check for several extensions
	if ((isExtensionActive('admins')) && ($table == 'admins')) {
		// Create email link for contacting admin in guest area
		$EMAIL = generateAdminEmailLink($email);
	} elseif ((isExtensionInstalledAndNewer('user', '0.3.3')) && ($table == 'user_data')) {
		// Create email link for contacting a member within admin area (or later in other areas, too?)
		$EMAIL = generateUserEmailLink($email);
	} elseif ((isExtensionActive('sponsor')) && ($table == 'sponsor_data')) {
		// Create email link to contact sponsor within admin area (or like the link above?)
		$EMAIL = generateSponsorEmailLink($email);
	}

	// Return email link
	return $EMAIL;
}

/**
 * Outputs an error message in a "fashioned way" to the user, by putting it into
 * a nice looking web page, if one of HTML or CSS output mode is active.
 *
 * Please use reportBug() instead of this function. reportBug() has more helpful
 * functionality like logging and admin notification (which you can configure
 * through your admin area).
 *
 * @param	$file		Function or file basename where the error came from
 * @param	$line		Line number where the error came from
 * @param	$message	Message which shall be output to web
 * @return	void
 */
function app_exit ($file, $line, $message) {
	// Check if Script is already dieing and not let it kill itself another 1000 times
	if (isset($GLOBALS['app_died'])) {
		// Script tried to kill itself twice
		die('[' . __FUNCTION__ . ':' . __LINE__ . ']: Script wanted to kill itself more than once! Raw message=' . $message . ', file/function=' . $file . ', line=' . $line);
	} // END - if

	// Make sure, that the script realy realy diese here and now
	$GLOBALS['app_died'] = TRUE;

	// Is this AJAX mode?
	if (isAjaxOutputMode()) {
		// Set content type as application/json
		setContentType('application/json');
	} else {
		// Set content type as text/html
		setContentType('text/html');
	}

	// Load header
	loadPageHeader();

	// Rewrite message for output
	$message = sprintf(
		getMessage('MAILER_HAS_DIED'),
		basename($file),
		$line,
		$message
	);

	// Is this AJAX mode again
	if (isAjaxOutputMode()) {
		// Load the message template
		$OUT = loadTemplate('ajax_app_exit_message', TRUE, $message);

		// Output it as JSON encoded
		outputHtml(encodeJson(array('reply_content' => urlencode(doFinalCompilation(compileRawCode($OUT))))));
	} else {
		// Load the message template
		loadTemplate('app_exit_message', FALSE, $message);
	}

	// Load footer
	loadPageFooter();
}

// Display parsing time and number of SQL queries in footer
function displayParsingTime () {
	// Is the timer started?
	if (!isset($GLOBALS['__start_time'])) {
		// Abort here
		return FALSE;
	} // END - if

	// Get end time
	$endTime = microtime(TRUE);

	// "Explode" both times
	$start = explode(' ', $GLOBALS['__start_time']);
	$end = explode(' ', $endTime);
	$runTime = $end[0] - $start[0];
	if ($runTime < 0) {
		$runTime = '0';
	} // END - if

	// Prepare output
	// @TODO This can be easily moved out after the merge from EL branch to this is complete
	$content = array(
		'run_time' => $runTime,
		'sql_time' => (getConfig('sql_time') * 1000),
	);

	// Load the template
	$GLOBALS['__page_footer'] .= loadTemplate('show_timings', TRUE, $content);

	// Is debug enabled?
	if (isDebugModeEnabled()) {
		// Log loading of total includes
		//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Loaded includes: ' . count($GLOBALS['inc_loaded']) . ', readable files: ' . count($GLOBALS['file_readable']));
	} // END - if
}

/**
 * Outputs an error message and backtrace to the user, by default a mail with
 * all relevant data is being mailed to the configured administrators.
 *
 * This function shall be used "publicly" because of logging, admin notification
 * and double-call prevention (see first if() block) instead of app_exit().
 * app_exit() is more a "private" function and will only output a bug message to
 * the user, no email and no logging.
 *
 * @param	$file		Function or file basename where the error came from
 * @param	$line		Line number where the error came from
 * @param	$sendEmail	Wether to send an email to all configured administrators
 * @return	void
 */
function reportBug ($file, $line, $message = '', $sendEmail = TRUE) {
	// Is this already called?
	if (isset($GLOBALS[__FUNCTION__])) {
		// Other backtrace
		print '[' . $file . ':' . $line . ':] ' . __FUNCTION__ . '() has already died! Message:' . $message . '<br />Backtrace:<pre>';
		debug_print_backtrace();
		die('</pre>');
	} // END - if

	// Set HTTP status to 500 (e.g. for AJAX requests)
	setHttpStatus('500 Internal Server Error');

	// Mark this function as called
	$GLOBALS[__FUNCTION__] = TRUE;

	// Init message
	$debug = '';

	// Is the optional message set?
	if (!empty($message)) {
		// Use and log it
		$debug = sprintf("Note: %s<br />\n",
			$message
		);

		// @TODO Add a little more infos here
		logDebugMessage($file, $line, strip_tags($message));
	} // END - if

	// Add output
	$debug .= 'Please report this bug at <a title="Direct link to the bug-tracker" href="http://bugs.mxchange.org" rel="external" target="_blank">http://bugs.mxchange.org</a> and include this whole message + logfile from <strong>' . str_replace(getPath(), '', getCachePath()) . 'debug.log</strong> in your report (you can now attach files).<br />Backtrace:<pre>';
	$debug .= debug_get_printable_backtrace();
	$debug .= '</pre>';
	$debug .= '<div class="para">Request-URI: ' . getRequestUri() . '</div>';
	$debug .= '<div class="para">Thank you for finding bugs.</div>';

	// Send an email? (e.g. not wanted for evaluation errors)
	if (($sendEmail === TRUE) && (!isInstaller()) && (isAdminRegistered())) {
		// Prepare content
		$content = array(
			'message'   => trim($message),
			'backtrace' => trim(debug_get_mailable_backtrace())
		);

		// Send email to webmaster
		sendAdminNotification('{--REPORT_BUG_SUBJECT--}', 'admin_report_bug', $content);
	} // END - if

	// Is there HTML/CSS/AJAX mode while debug-mode is enabled?
	if (((isDebugModeEnabled()) || (isInstaller())) && ((isHtmlOutputMode()) || (isCssOutputMode()) || (isAjaxOutputMode()))) {
		// And abort here
		app_exit($file, $line, $debug);
	} elseif (isAjaxOutputMode()) {
		// Load include (again?)
		loadIncludeOnce('inc/ajax-functions.php');

		// Init AJAX
		initAjax();

		// Set HTTP status to 500 again
		setHttpStatus('500 Internal Server Error');

		// Is AJAX output mode, then output message as JSON
		setAjaxReplyContent($debug);

		// Send it out to browser
		sendAjaxContent(TRUE);

		// Include footer
		loadPageFooter();
	} else {
		// Raw/image output mode and all other modes doesn't work well with text ...
		die();
	}
}

// Compile characters which are allowed in URLs
function compileUriCode ($code, $simple = TRUE) {
	// Trim code
	$test = trim($code);

	// Is it empty?
	if (empty($test)) {
		// Then abort here and return the original code
		return $code;
	} // END - if

	// Compile these by default
	$charsCompile = array(
		'from' => array(
			'{DOT}',
			'{SLASH}',
			'{QUOT}',
			'{DOLLAR}',
			'{OPEN_ANCHOR}',
			'{CLOSE_ANCHOR}',
			'{OPEN_SQR}',
			'{CLOSE_SQR}',
			'{PER}'
		),
		'to' => array(
			'.',
			'/',
			chr(39),
			'$',
			'(',
			')',
			'[',
			']',
			'%'
		)
	);

	// Compile constants
	if ($simple === FALSE) {
		// Add more 'from'
		array_unshift($charsCompile['from'], '{--', '--}');

		// Add more 'to'
		array_unshift($charsCompile['to'], '".', '."');
	} // END - if

	// Compile QUOT and other non-HTML codes
	$code = str_replace($charsCompile['from'], $charsCompile['to'], $code);
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'code=' . $code);

	// Return compiled code
	return $code;
}

// Handle message codes from URL
function handleCodeMessage () {
	// Is 'code' set?
	if (isGetRequestElementSet('code')) {
		// Default extension is 'unknown'
		$ext = 'unknown';

		// Is extension given?
		if (isGetRequestElementSet('ext')) {
			$ext = getRequestElement('ext');
		} // END - if

		// Convert the 'code' parameter from URL to a human-readable message
		$message = getMessageFromErrorCode(getRequestElement('code'));

		// Load message template
		loadTemplate('message', FALSE, $message);
	} // END - if
}

// Generates a 'extension foo out-dated' message
function generateExtensionOutdatedMessage ($ext_name, $ext_ver) {
	// Is the extension empty?
	if (empty($ext_name)) {
		// This should not happen
		reportBug(__FUNCTION__, __LINE__, 'Parameter ext is empty. This should not happen.');
	} // END - if

	// Default message
	$message = '{%message,EXTENSION_PROBLEM_EXTENSION_OUTDATED=' . $ext_name . '%}';

	// Is an admin logged in?
	if (isAdmin()) {
		// Then output admin message
		$message = sprintf(getMessage('ADMIN_EXTENSION_PROBLEM_EXTENSION_INACTIVE'), $ext_name, $ext_name, $ext_ver);
	} // END - if

	// Return prepared message
	return $message;
}

// Generates a 'extension foo inactive' message
function generateExtensionInactiveMessage ($ext_name) {
	// Is the extension empty?
	if (empty($ext_name)) {
		// This should not happen
		reportBug(__FUNCTION__, __LINE__, 'Parameter ext is empty. This should not happen.');
	} // END - if

	// Default message
	$message = '{%message,EXTENSION_PROBLEM_EXTENSION_INACTIVE=' . $ext_name . '%}';

	// Is an admin logged in?
	if (isAdmin()) {
		// Then output admin message
		$message = '{%message,ADMIN_EXTENSION_PROBLEM_EXTENSION_INACTIVE=' . $ext_name . '%}';
	} // END - if

	// Return prepared message
	return $message;
}

// Generates a 'extension foo not installed' message
function generateExtensionNotInstalledMessage ($ext_name) {
	// Is the extension empty?
	if (empty($ext_name)) {
		// This should not happen
		reportBug(__FUNCTION__, __LINE__, 'Parameter ext is empty. This should not happen.');
	} // END - if

	// Default message
	$message = '{%message,EXTENSION_PROBLEM_EXTENSION_NOT_INSTALLED=' . $ext_name . '%}';

	// Is an admin logged in?
	if (isAdmin()) {
		// Then output admin message
		$message = '{%message,ADMIN_EXTENSION_PROBLEM_EXTENSION_NOT_INSTALLED=' . $ext_name . '%}';
	} // END - if

	// Return prepared message
	return $message;
}

// Generates a message depending on if the extension is not installed or not
// just activated
function generateExtensionInactiveNotInstalledMessage ($ext_name) {
	// Init message
	$message = '';

	// Is the extension not installed or just deactivated?
	switch (isExtensionInstalled($ext_name)) {
		case TRUE; // Deactivated!
			$message = generateExtensionInactiveMessage($ext_name);
			break;

		case FALSE; // Not installed!
			$message = generateExtensionNotInstalledMessage($ext_name);
			break;

		default: // Should not happen!
			logDebugMessage(__FUNCTION__, __LINE__, sprintf('Invalid state of extension %s detected.', $ext_name));
			$message = sprintf('Invalid state of extension %s detected.', $ext_name);
			break;
	} // END - switch

	// Return the message
	return $message;
}

// Print code with line numbers
function lineNumberCode ($code)    {
	// By default copy the code
	$codeE = $code;

	if (!is_array($code)) {
		// We need an array, so try it with the new-line character
		$codeE = explode(PHP_EOL, $code);
	} // END - if

	$count_lines = count($codeE);

	$r = 'Line | Code:<br />';
	foreach ($codeE as $line => $c) {
		$r .= '<div class="line"><span class="linenum">';
		if ($count_lines == 1) {
			$r .= 1;
		} else {
			$r .= ($line == ($count_lines - 1)) ? '' : ($line+1);
		}
		$r .= '</span>|';

		// Add code
		$r .= '<span class="linetext">' . encodeEntities($c) . '</span></div>';
	} // END - foreach

	return '<div class="code">' . $r . '</div>';
}

// Determines the right page title
function determinePageTitle () {
	// Init page title
	$pageTitle = '';

	// Config and database connection valid?
	if ((isConfigLocalLoaded()) && (isConfigurationLoaded()) && (isSqlLinkUp()) && (isExtensionInstalledAndNewer('sql_patches', '0.1.6'))) {
		// Title decoration enabled?
		if ((isTitleDecorationEnabled()) && (getTitleLeft() != '')) {
			$pageTitle .= '{%config,trim=title_left%} ';
		} // END - if

		// Is there an extra title?
		if (isExtraTitleSet()) {
			// Then prepend it
			$pageTitle .= '{%pipe,getExtraTitle%} by ';
		} // END - if

		// Add main title
		$pageTitle .= '{?MAIN_TITLE?}';

		// Add title of module? (middle decoration will also be added!)
		if ((isModuleTitleEnabled()) || ((!isWhatSet()) && (!isActionSet())) || (getModule() == 'admin')) {
			$pageTitle .= ' {%config,trim=title_middle%} {DQUOTE} . getModuleTitle(getModule()) . {DQUOTE}';
		} // END - if

		// Get menu mode from module
		$menuMode = getMenuModeFromModule();

		// Add middle part (always in admin area!)
		if ((!empty($menuMode)) && ((isWhatTitleEnabled()) || ($menuMode == 'admin'))) {
			$pageTitle .= ' {%config,trim=title_middle%} ' . getTitleFromMenu($menuMode, getWhat());
		} // END - if

		// Add title decorations? (right)
		if ((isTitleDecorationEnabled()) && (getTitleRight() != '')) {
			$pageTitle .= ' {%config,trim=title_right%}';
		} // END - if
	} elseif ((isInstalled()) && (isAdminRegistered())) {
		// Installed, admin registered but no ext-sql_patches
		$pageTitle = '[-- {?MAIN_TITLE?} - {%pipe,getModule,getModuleTitle%} --]';
	} elseif ((isInstalled()) && (!isAdminRegistered())) {
		// Installed but no admin registered
		$pageTitle = '{--INSTALLER_OF_MAILER_NO_ADMIN--}';
	} elseif ((!isInstalled()) || (!isAdminRegistered())) {
		// Installation mode
		$pageTitle = '{--INSTALLER_OF_MAILER--}';
	} else {
		// Configuration not found
		$pageTitle = '{--NO_CONFIG_FOUND_TITLE--}';

		// Do not add the fatal message in installation mode
		if ((!isInstalling()) && (!isConfigurationLoaded())) {
			// Please report this
			reportBug(__FUNCTION__, __LINE__, 'No configuration data found!');
		} // END - if
	}

	// Return title
	return decodeEntities($pageTitle);
}

// Checks whethere there is a cache file there. This function is cached.
function isTemplateCached ($prefix, $template) {
	// Is there cached this result?
	if (!isset($GLOBALS['template_cache'][$prefix][$template])) {
		// Generate FQFN
		$FQFN = generateCacheFqfn($prefix, $template);

		// Is it there?
		$GLOBALS['template_cache'][$prefix][$template] = isFileReadable($FQFN);
	} // END - if

	// Return it
	return $GLOBALS['template_cache'][$prefix][$template];
}

// Flushes non-flushed template cache to disk
function flushTemplateCache ($prefix, $template, $eval) {
	// Is this cache flushed?
	if ((isDebugTemplateCacheEnabled() === FALSE) && (isTemplateCached($prefix, $template) === FALSE) && ($eval != '404')) {
		// Generate FQFN
		$FQFN = generateCacheFqfn($prefix, $template);

		// Compile code another round for better performance and preserve $ signs
		$eval = str_replace(array(chr(92), '{DOLLAR}', '{BACK}', '{CONTENT}'), array('', '$', chr(92), '$content'), compileCode(str_replace(array('$content', chr(92)), array('{CONTENT}', '{BACK}'), $eval)));

		// Is this a XML template?
		if ($prefix == 'xml') {
			// Compact only XML templates as emails needs new-line characters and HTML may contain required "comments"
			$eval = compactContent($eval);
		} // END - if

		// And flush it
		writeToFile($FQFN, '<?php ' . $eval . ' ?>', TRUE);
	} // END - if
}

// Reads a template cache
function readTemplateCache ($prefix, $template, $content) {
	// Check it again
	if ((isDebugTemplateCacheEnabled()) || (!isTemplateCached($prefix, $template))) {
		// This should not happen
		reportBug(__FUNCTION__, __LINE__, 'Wether debugging of template cache is enabled or template ' . $template . ' is not cached while expected.');
	} // END - if

	// Is it cached?
	if (!isset($GLOBALS['template_eval'][$prefix][$template])) {
		// Generate FQFN
		$FQFN = generateCacheFqfn($prefix, $template);

		/*
		 * And read from it.
		 *
		 * WARNING: Do not replace this include() call with loadInclude() as it
		 * would hide local variables away which is here required to make this
		 * work.
		 */
		include($FQFN);

		// Is the template cache valid?
		if (!isset($templateContent)) {
			// Please clear your cache!
			reportBug(__FUNCTION__, __LINE__, 'Template ' . $template . ' uses old structure. Please delete all template cache files and reload.');
		} // END - if
	} // END - if

	// And return it
	return $templateContent;
}

// Escapes quotes (default is only double-quotes)
function escapeQuotes ($str, $single = FALSE) {
	// Should we escape all?
	if ($single === TRUE) {
		// Escape all (including null)
		$str = addslashes($str);
	} else {
		// Replace all chars at once
		$str = str_replace(array("\\'", '"', "\\\\"), array(chr(39), "\\\"", chr(92)), $str);
	}

	// Return the escape'd string
	return $str;
}

// Escapes the JavaScript code, prevents \r and \n becoming char 10/13
function escapeJavaScriptQuotes ($str) {
	// Replace all double-quotes and secure back-ticks
	$str = str_replace(array(chr(92), '"'), array('{BACK}', '\"'), $str);

	// Return it
	return $str;
}

// Send out mails depending on the 'mod/modes' combination
// @TODO Lame description for this function
function sendModeMails ($mod, $modes) {
	// Init user data
	$content = array ();

	// Load hash
	if (fetchUserData(getMemberId())) {
		// Extract salt from cookie
		$salt = substr(getSession('u_hash'), 0, -40);

		// Now let's compare passwords
		$hash = encodeHashForCookie(getUserData('password'));

		// Does the hash match or should we change it?
		if (($hash == getSession('u_hash')) || (postRequestElement('password1') == postRequestElement('password2'))) {
			// Load the data
			$content = getUserDataArray();

			// Clear/init the content variable
			$content['message'] = '';

			// Which mail?
			// @TODO Move this in a filter
			switch ($mod) {
				case 'mydata':
					foreach ($modes as $mode) {
						switch ($mode) {
							case 'normal': break; // Do not add any special lines
							case 'email': // Email was changed!
								$content['message'] = '{--MEMBER_CHANGED_EMAIL--}' . ': ' . postRequestElement('old_email') . PHP_EOL;
								break;

							case 'password': // Password was changed
								$content['message'] = '{--MEMBER_CHANGED_PASSWORD--}' . PHP_EOL;
								break;

							default:
								logDebugMessage(__FUNCTION__, __LINE__, sprintf('Unknown mode %s detected.', $mode));
								$content['message'] = '{%message,MEMBER_UNKNOWN_MODE=' .  $mode . '%}' . PHP_EOL . PHP_EOL;
								break;
						} // END - switch
					} // END - foreach

					if (isExtensionActive('country')) {
						// Replace code with description
						$content['country'] = generateCountryInfo(postRequestElement('country_code'));
					} // END - if

					// Merge content with data from POST
					$content = merge_array($content, postRequestArray());

					// Load template
					$message = loadEmailTemplate('member_mydata_notify', $content, getMemberId());

					if (isAdminNotificationEnabled()) {
						// The admin needs to be notified about a profile change
						$message_admin = 'admin_mydata_notify';
						$subjectAdmin   = '{--ADMIN_CHANGED_DATA_SUBJECT--}';
					} else {
						// No mail to admin
						$message_admin = '';
						$subjectAdmin  = '';
					}

					// Set subject lines
					$subjectMember = '{--MEMBER_CHANGED_DATA--}';

					// Output success message
					$content['message'] = '<span class="message">{--MEMBER_MYDATA_MAIL_SENT--}</span>';
					break;

				default: // Unsupported module!
					logDebugMessage(__FUNCTION__, __LINE__, sprintf('Unsupported module %s detected.', $mod));
					$content['message'] = '<span class="bad">{%message,UNKNOWN_MODULE=' . $mod . '%}</span>';
					break;
			} // END - switch
		} else {
			// Passwords mismatch
			$content['message'] = '<span class="bad">{--MEMBER_PASSWORD_ERROR--}</span>';
		}
	} else {
		// Could not load profile
		$content['message'] = '<span class="bad">{--MEMBER_CANNOT_LOAD_PROFILE--}</span>';
	}

	// Send email to user if required
	if ((!empty($subjectMember)) && (!empty($message)) && (!empty($content['userid']))) {
		// Send member mail
		sendEmail($content['userid'], $subjectMember, $message);
	} // END - if

	// Send only if no other error has occured
	if ((!empty($subjectAdmin)) && (!empty($message_admin)) && (isAdminNotificationEnabled())) {
		// Send admin mail
		sendAdminNotification($subjectAdmin, $message_admin, $content, getMemberId());
	} elseif (isAdminNotificationEnabled()) {
		// Cannot send mails to admin!
		$content['message'] = '{--CANNOT_SEND_ADMIN_MAILS--}';
	} else {
		// No mail to admin
		$content['message'] = '<span class="message">{--MEMBER_MYDATA_MAIL_SENT--}</span>';
	}

	// Load template
	displayMessage($content['message']);
}

// Generates a 'selection box' from given array
function generateSelectionBoxFromArray ($options, $name, $optionKey, $optionContent = '', $extraName = '', $templateName = '', $default = NULL, $nameElement = '', $allowNone = FALSE, $useDefaultAsArray = FALSE) {
	// options must be an array
	assert(is_array($options));

	// Default is empty
	$addKey = '';

	// Use default value as array key?
	if ($useDefaultAsArray === TRUE) {
		// Then set it
		$addKey = '[' . convertNullToZero($default) . ']';
	} // END - if

	// Start the output
	$OUT = '<select name="' . $name . $addKey . '" size="1" class="form_select">
<option value="X" disabled="disabled">{--PLEASE_SELECT--}</option>';

	// Allow none?
	if ($allowNone === TRUE) {
		// Then add it
		$OUT .= '<option value="0">{--SELECT_NONE--}</option>';
	} // END - if

	// Walk through all options
	foreach ($options as $option) {
		// Default 'default' is not set
		$option['default'] = '';

		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'name=' . $name . ',default[' . gettype($default) . ']=' . $default . ',optionKey[' . gettype($optionKey) . ']=' . $optionKey);
		// Is default value same as given value?
		if ((!is_null($default)) && (isset($option[$optionKey])) && ($default == $option[$optionKey])) {
			// Then set default
			$option['default'] = ' selected="selected"';
		} // END - if

		// Is 'nameElement' set?
		if ((!empty($nameElement)) && (isset($option[$nameElement]))) {
			// Then set this as extraName, but lower-case
			$extraName = '_' . strtolower($option[$nameElement]);
		} // END - if

		// Add the <option> entry from ...
		if (empty($optionContent)) {
			// Is a template name given?
			if (empty($templateName)) {
				// ... $name template
				$OUT .= loadTemplate('select_' . $name . $extraName . '_option', TRUE, $option);
			} else {
				// ... $templateName template
				$OUT .= loadTemplate('select_' . $templateName . $extraName . '_option', TRUE, $option);
			}
		} else {
			// ... direct HTML code
			$OUT .= '<option value="' . $option[$optionKey] . '">' . $option[$optionContent] . '</option>';
		}
	} // END - foreach

	// Finish selection box
	$OUT .= '</select>';

	// Prepare output
	$content = array(
		'selection_box' => $OUT,
	);

	// Load template and return it
	if (empty($templateName)) {
		// Use name from $name + $extraName
		return loadTemplate('select_' . $name . $extraName . '_box', TRUE, $content);
	} else {
		// Use name from $templateName + $extraName
		return loadTemplate('select_' . $templateName . $extraName . '_box', TRUE, $content);
	}
}

// Prepares the header for HTML output
function loadHtmlHeader () {
	/*
	 * Run two filters:
	 * 1.) pre_page_header (mainly loads the page_header template and includes
	 *     meta description)
	 */
	runFilterChain('pre_page_header');

	/*
	 * Here can be something be added, but normally one of the two filters
	 * around this line should do the job for you.
	 */

	/*
	 * 2.) post_page_header (mainly to load stylesheet, extra JavaScripts and
	 *     to close the head-tag)
	 * Include more header data here
	 */
	runFilterChain('post_page_header');
}

// Adds page header and footer to output array element
function addPageHeaderFooter () {
	// Init output
	$OUT = '';

	// Add them all together. This is maybe to simple
	foreach (array('__page_header', '__output', '__page_footer') as $pagePart) {
		// Add page part if set
		if (isset($GLOBALS[$pagePart])) {
			$OUT .= $GLOBALS[$pagePart];
		} // END - if
	} // END - foreach

	// Transfer $OUT to '__output'
	$GLOBALS['__output'] = $OUT;
}

// Generates meta description for current module and 'what' value
function generateMetaDescriptionCode () {
	// Only include from guest area and if ext-sql_patches has correct version
	if ((getModule() == 'index') && (isExtensionInstalledAndNewer('sql_patches', '0.1.6'))) {
		// Output it directly
		$GLOBALS['__page_header'] .= '<meta name="description" content="' . '{?MAIN_TITLE?} ' . trim(getConfig('title_middle')) . ' ' . getTitleFromMenu('guest', getWhat()) . '" />';
	} // END - if

	// Initialize referral system
	initReferralSystem();
}

// Generates an FQFN for template cache from the given template name
function generateCacheFqfn ($prefix, $template) {
	// Is this cached?
	if (!isset($GLOBALS['template_cache_fqfn'][$prefix][$template])) {
		// Generate the FQFN
		$GLOBALS['template_cache_fqfn'][$prefix][$template] = sprintf(
			'%s%s_compiled/%s/%s%s',
			getPath(),
			getCachePath(),
			$prefix,
			$template,
			getCacheExtension()
		);
	} // END - if

	// Return it
	return $GLOBALS['template_cache_fqfn'][$prefix][$template];
}

// "Fixes" null or empty string to count of dashes
function fixNullEmptyToDashes ($str, $num) {
	// Use str as default
	$return = $str;

	// Is it empty?
	if ((is_null($str)) || (trim($str) == '')) {
		// Set it
		$return = str_repeat('-', $num);
	} // END - if

	// Return final string
	return $return;
}

// Translates the "pool type" into human-readable
function translatePoolType ($type) {
	// Return "translation"
	return sprintf('{--POOL_TYPE_%s--}', strtoupper($type));
}

// "Translates" given time unit
function translateTimeUnit ($unit) {
	// Default is unknown
	$message = '{%message,TIME_UNIT_UNKNOWN=' . $unit . '%}';

	// "Detect" it
	if (!isset($GLOBALS['time_units'][$unit])) {
		// Not found
		logDebugMessage(__FUNCTION__, __LINE__, 'Unknown time unit ' . $unit . ' detected.');
	} else {
		// Translate it with generic function
		$message = translateGeneric('TIME_UNIT' , $GLOBALS['time_units'][$unit]);
	}

	// Return message
	return $message;
}

// Displays given message in admin_settings_saved template
function displayMessage ($message) {
	// Call inner function
	outputHtml(returnMessage($message));
}

// Returns given message in admin_settings_saved template
function returnMessage ($message) {
	// Load the template
	return loadTemplate('admin_settings_saved', TRUE, $message);
}

// Displays given error message in admin_settings_unsaved template
function displayErrorMessage ($message) {
	// Load the template
	outputHtml(returnErrorMessage($message));
}

// Displays given error message in admin_settings_unsaved template
function returnErrorMessage ($message) {
	// Load the template
	return loadTemplate('admin_settings_unsaved', TRUE, $message);
}

// Generates a selection box for (maybe) given gender
function generateGenderSelectionBox ($selectedGender = '', $fieldName = 'gender') {
	// Start the HTML code
	$out  = '<select name="' . $fieldName . '" size="1" class="form_select">';

	// Add options
	$out .= generateOptions(
		'/ARRAY/',
		array(
			'M',
			'F',
			'C'
		), array(
			'{--GENDER_M--}',
			'{--GENDER_F--}',
			'{--GENDER_C--}'
		),
		$selectedGender
	);

	// Finish HTML code
	$out .= '</select>';

	// Return the code
	return $out;
}

// Generates a selection box for given default value
function generateTimeUnitSelectionBox ($defaultUnit, $fieldName, $unitArray) {
	// Init variables
	$messageIds = array();

	// Generate message id array
	foreach ($unitArray as $unit) {
		// "Translate" it
		array_push($messageIds, '{%pipe,translateTimeUnit=' . $unit . '%}');
	} // END - foreach

	// Start the HTML code
	$out = '<select name="' . $fieldName . '" size="1" class="form_select">';

	// Add options
	$out .= generateOptions('/ARRAY/', $unitArray, $messageIds, $defaultUnit);

	// Finish HTML code
	$out .= '</select>';

	// Return the code
	return $out;
}

// Function to add style tag (whether display:none/block)
function addStyleMenuContent ($menuMode, $mainAction, $action) {
	// Is there foo_menu_javascript enabled?
	if ((!isConfigEntrySet($menuMode . '_menu_javascript')) || (getConfig($menuMode . '_menu_javascript') == 'N')) {
		// Silently abort here, not enabled
		return '';
	} // END - if

	// Is action=mainAction?
	if ($action == $mainAction) {
		// Add "menu open" style
		return ' style="display:block"';
	} else {
		return ' style="display:none"';
	}
}

// Function to add onclick attribute
function addJavaScriptMenuContent ($menuMode, $mainAction, $action, $what) {
	// Is there foo_menu_javascript enabled?
	if ((!isConfigEntrySet($menuMode . '_menu_javascript')) || (getConfig($menuMode . '_menu_javascript') == 'N')) {
		// Silently abort here, not enabled
		return '';
	} // END - if

	// Prepare output
	$OUT = ' onclick="return changeMenuFoldState(' . $menuMode . ', ' . $mainAction . ', ' . $action . ', ' . $what . ')';

	// Return output
	return $OUT;
}

// Tries to anonymize some sensitive data (e.g. IP address, user agent, referrer, etc.)
function anonymizeSensitiveData ($data) {
	// Trim it
	$data = trim($data);

	// Is it empty?
	if (empty($data)) {
		// Then add three dashes
		$data = '---';
	} elseif (isUrlValid($data)) {
		// Is a referrer, so is it black-listed?
		if (isAdmin()) {
			// Is admin, has always priority
			$data = '[<a href="{%pipe,generateFrametesterUrl=' . $data . '%}" target="_blank">{--ADMIN_TEST_URL--}</a>]';
		} elseif ((isExtensionActive('blacklist')) && (isUrlBlacklisted($data))) {
			// Yes, so replace it with text
			$data = '<em>{--URL_IS_BLACKLISTED--}</em>';
		} else {
			// A  member is viewing this referral URL
			$data = '[<a href="{%pipe,generateDereferrerUrl=' . $data . '%}" target="_blank">{--MEMBER_TEST_URL--}</a>]';
		}
	} elseif (isIp4AddressValid($data)) {
		// Is an IPv4 address
		$ipArray = explode('.', $data);

		// Only display first 2 octets
		$data = $ipArray[0] . '.' . $ipArray[1] . '.?.?';
	} else {
		// Generic data
		$data = '<em>{--DATA_IS_HIDDEN--}</em>';
	}

	// Return it (hopefully) anonymized
	return $data;
}

/**
 * Removes all comments, tabs and new-line characters to compact the content
 *
 * @param	$uncompactedContent		The uncompacted content
 * @return	$compactedContent		The compacted content
 */
function compactContent ($uncompactedContent) {
	// First, remove all tab/new-line/revert characters
	$compactedContent = str_replace(chr(9), '', str_replace(PHP_EOL, '', str_replace(chr(13), '', $uncompactedContent)));

	// Make a space after >
	$compactedContent = str_replace(array('>', '  '), array('> ', ' '), $compactedContent);

	// Then regex all comments like <!-- //--> away
	preg_match_all('/<!--[\w\W]*?(\/\/){0,1}-->/', $compactedContent, $matches);

	// Do we have entries?
	if (isset($matches[0][0])) {
		// Remove all
		foreach ($matches[0] as $match) {
			// Remove the match
			$compactedContent = str_replace($match, '', $compactedContent);
		} // END - foreach
	} // END - if

	// Return compacted content
	return $compactedContent;
}

//-----------------------------------------------------------------------------
//                     Template helper functions for EL code
//-----------------------------------------------------------------------------

// Color-switch helper function
function doTemplateColorSwitch ($templateName, $clear = FALSE, $return = TRUE) {
	// Is it there?
	if (!isset($GLOBALS['color_switch'][$templateName])) {
		// Initialize it
		initTemplateColorSwitch($templateName);
	} elseif ($clear === FALSE) {
		// Switch color if called from loadTemplate()
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'SWITCH:' . $templateName);
		$GLOBALS['color_switch'][$templateName] = 3 - $GLOBALS['color_switch'][$templateName];
	}

	// Return CSS class name
	if ($return === TRUE) {
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'RETURN:' . $templateName . '=' . $GLOBALS['color_switch'][$templateName]);
		return 'switch_sw' . $GLOBALS['color_switch'][$templateName];
	} // END - if
}

// Helper function for extension registration link
function doTemplateExtensionRegistrationLink ($templateName, $clear, $ext_name) {
	// Default is all non-productive
	$OUT = '<div style="cursor:help" title="{%message,ADMIN_EXTENSION_IS_NON_PRODUCTIVE_LINK_TITLE=' . $ext_name . '%}">{--ADMIN_EXTENSION_IS_NON_PRODUCTIVE_LINK--}</div>';

	// Is the given extension non-productive?
	if (isExtensionDeprecated($ext_name)) {
		// Is deprecated
		$OUT = '<span title="{--ADMIN_EXTENSION_IS_DEPRECATED_TITLE--}">---</span>';
	} elseif (isExtensionProductive($ext_name)) {
		// Productive code
		$OUT = '<a title="{--ADMIN_TASK_REGISTER_EXTENSION_TITLE--}" href="{%url=modules.php?module=admin&amp;what=extensions&amp;register_ext=' . $ext_name . '%}">{--ADMIN_TASK_REGISTER_EXTENSION--}</a>';
	}

	// Return code
	return $OUT;
}

// Helper function to create bonus mail admin links
function doTemplateAdminBonusMailLinks ($templateName, $clear, $bonusId) {
	// Call the inner function
	return generateAdminMailLinks('bonus', $bonusId);
}

// Helper function to create member mail admin links
function doTemplateAdminMemberMailLinks ($templateName, $clear, $mailId) {
	// Call the inner function
	return generateAdminMailLinks('normal', $mailId);
}

// Helper function to create a selection box for YES/NO configuration entries
function doTemplateConfigurationYesNoSelectionBox ($templateName, $clear, $configEntry) {
	// Default is a "missing entry" warning
	$OUT = '<div class="bad" style="cursor:help" title="{%message,ADMIN_CONFIG_ENTRY_MISSING=' . $configEntry . '%}">!' . $configEntry . '!</div>';

	// Generate the HTML code
	if (isConfigEntrySet($configEntry)) {
		// Configuration entry is found
		$OUT = '<select name="' . $configEntry . '" class="form_select" size="1">
{%config,generateYesNoOptions=' . $configEntry . '%}
</select>';
	} // END - if

	// Return it
	return $OUT;
}

// Helper function to create a selection box for YES/NO form fields
function doTemplateYesNoSelectionBox ($templateName, $clear, $formField) {
	// Generate the HTML code
	$OUT = '<select name="' . $formField . '" class="form_select" size="1">
{%pipe,generateYesNoOptions%}
</select>';

	// Return it
	return $OUT;
}

// Helper function to create a selection box for YES/NO form fields, by NO is default
function doTemplateNoYesSelectionBox ($templateName, $clear, $formField) {
	// Generate the HTML code
	$OUT = '<select name="' . $formField . '" class="form_select" size="1">
{%pipe,generateYesNoOptions=N%}
</select>';

	// Return it
	return $OUT;
}

// Helper function to add extra content for guest area (module=index and others)
function doTemplateGuestFooterExtras ($templateName, $clear) {
	// Init filter data
	$filterData = array(
		// Name of used template
		'template' => $templateName,
		// Target array for gathered data
		'__data'   => array(),
		// Where the HTML output will go
		'__output' => '',
	);

	// Run the filter chain
	$filterData = runFilterChain('guest_footer_extras', $filterData);

	// Return output
	return $filterData['__output'];
}

// Helper function to add extra content for member area (module=login)
function doTemplateMemberFooterExtras ($templateName, $clear) {
	// Is a member logged in?
	if (!isMember()) {
		// This shall not happen
		reportBug(__FUNCTION__, __LINE__, 'Please use this template helper only for logged-in members.');
	} // END - if

	// Init filter data
	$filterData = array(
		// Current user's id number
		'userid'   => getMemberId(),
		// Name of used template
		'template' => $templateName,
		// Target array for gathered data
		'__data'   => array(),
		// Where the HTML output will go
		'__output' => '',
	);

	// Run the filter chain
	$filterData = runFilterChain('member_footer_extras', $filterData);

	// Return output
	return $filterData['__output'];
}

/**
 * Helper function to determine whether current userid is set, if none is set,
 * return a zero, else an EL code is being returned as of this function is used
 * only in templates.
 *
 * @param	$templateName	Name of template (unused)
 * @param	$clear			Wether to clear something (unused)
 * @return	$userId			Wether zero or EL code snippet
 */
function doTemplateUserId ($templateName, $clear) {
	// By default no userid is set
	$userId = '0';

	// Is there a user id currently set?
	if (isCurrentUserIdSet()) {
		// Then get the current user id
		$userId = getCurrentUserId();
	} // END - if

	// Return it
	return $userId;
}

// Template helper function to generate "Terms&Conditions" link (EL code again)
function doTemplateGetTermsConditionsLink ($templateName, $clear) {
	/*
	 * Use default link by default ;-) This link, however, will become
	 * deprecated once ext-terms is rolled out.
	 */
	$linkCode = '{%url=modules.php?module=index&amp;what=agb%}';

	// Is ext-terms installed?
	if (isExtensionInstalled('terms')) {
		// Then use that link (only 'what' has changed)
		$linkCode = '{%url=modules.php?module=index&amp;what=terms%}';
	} // END - if

	// Return link (EL) code
	return $linkCode;
}

// Template helper function to create selection box for "locked points mode"
function doTemplatePointsLockedModeSelectionBox ($templateName, $clear = FALSE, $default = NULL) {
	// Init array
	$lockedModes = array(
		0 => array('mode' => 'LOCKED'),
		1 => array('mode' => 'UNLOCKED'),
	);

	// Handle it over to generateSelectionBoxFromArray()
	$content = generateSelectionBoxFromArray($lockedModes, 'points_locked_mode', 'mode', '', '', '', $default);

	// Return prepared content
	return $content;
}

// Template helper function to create selection box for payment method
function doTemplatePointsPaymentMethodSelectionBox ($templateName, $clear = FALSE, $default = NULL) {
	// Init array
	$paymentMethods = array(
		0 => array('method' => 'DIRECT'),
		1 => array('method' => 'REFERRAL'),
	);

	// Handle it over to generateSelectionBoxFromArray()
	$content = generateSelectionBoxFromArray($paymentMethods, 'points_payment_method', 'method', '', '', '', $default);

	// Return prepared content
	return $content;
}

// Template helper function to create a deferrer code if URL is not empty
function doTemplateDereferrerUrl ($templateName, $clear = FALSE, $url = NULL) {
	// Is the URL not NULL and not empty?
	if ((!is_null($url)) && (!empty($url))) {
		// Set HTML with EL code
		$url = '<a href="{%pipe,generateDereferrerUrl=' . $url . '%}" rel="external" target="_blank">{--ADMIN_TEST_URL--}</a>';
	} // END - if

	// Return URL (or content) or dashes if empty
	return fixEmptyContentToDashes($url);
}

// Load another template and return its content
function doTemplateLoadTemplate ($templateName, $clear = FALSE, $theTemplate, $content = array()) {
	// Load "the" template
	return loadTemplate($theTemplate, TRUE, $content);
}

// Output HTML code for favicon.ico, if found
function doTemplateMetaFavIcon ($templateName, $clear = FALSE) {
	// Default is not found
	$out = '';

	// Check all common extensions
	foreach (array('ico', 'gif', 'png') as $extension) {
		// Is the file there?
		if (isFileReadable(getPath() . 'favicon.' . $extension)) {
			// Then use this and abort
			$out = '<link rel="shortcut icon" href="{%url=favicon.' . $extension . '%}" type="image/' . $extension . '" />';
			break;
		} // END - if
	} // END - while

	// Return code
	return $out;
}

// Helper function to display referral id or hide it depending on settings
function doTemplateDisplayReferralIdContent ($template, $clear = FALSE) {
	// Ddisplay the refid or make it editable?
	if (isDisplayRefidEnabled()) {
		// Load "hide" form template
		$out = loadTemplate('guest_register_refid_hide', TRUE);
	} else {
		// Load template to enter it
		$out = loadTemplate('guest_register_refid', TRUE);
	}

	// Return code
	return $out;
}

// "Getter" for template base path
function getTemplateBasePath ($part) {
	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$part])) {
		// "Build" path
		$GLOBALS[__FUNCTION__][$part] = sprintf('%stemplates/%s/%s', getPath(), getLanguage(), $part);
	} // END - if

	// Return cache
	return $GLOBALS[__FUNCTION__][$part];
}

// Removes comments with @DEPRECATED
function removeDeprecatedComment ($output) {
	// Explode it into pieces ... ;)
	$lines = explode(chr(10), $output);

	// Walk through all
	$return = '';
	foreach ($lines as $line) {
		// Is there a @DEPRECATED?
		if (isInString('@DEPRECATED', $line)) {
			// Ignore this line
			continue;
		} // END - if

		// Add it
		$return .= $line . chr(13);
	} // END - foreach

	// Returned cleaned content
	return $return;
}

// Generates a selection box suitable for e.g. birthdays: day, month and year
function generateDayMonthYearSelectionBox ($day, $month, $year) {
 	// This depends on selected language
	switch (getLanguage()) {
		case 'de': // German date format
			$content = addSelectionBox('da', $day) . addSelectionBox('mo', $month) . addSelectionBox('ye', $year);
			break;

		default: // Default is the US date format... :)
			$content = addSelectionBox('mo', $month) . addSelectionBox('da', $day) . addSelectionBox('ye', $year);
			break;
	} // END - switch

	// Return content
	return $content;
}

// Loads page header
function loadPageHeader () {
	// Init header
	$GLOBALS['__page_header'] = '';

	// Is the header already sent?
	if (($GLOBALS['__header_sent'] != 1) && ($GLOBALS['__header_sent'] != 2)) {
		// Set default HTTP status to "200 OK"
		setHttpStatus('200 OK');

		// If not in CSS mode generate the header
		if ((!isCssOutputMode()) && (!isAjaxOutputMode()) && (!isImageOutputMode())) {
			// Prepare the header for HTML output
			loadHtmlHeader();
		} // END - if

		// Closing HEAD tag
		if ($GLOBALS['__header_sent'] == '0') {
			$GLOBALS['__header_sent'] = 1;
		} // END - if
	} // END - if

	// Add BODY tag or not?
	// @TODO Find a way to not use direct module comparison
	if ((!isCssOutputMode()) && (!isRawOutputMode()) && ($GLOBALS['__header_sent'] == 1) && (getModule() != 'frametester') && (!isFramesetModeEnabled())) {
		loadTemplate('page_body');
		$GLOBALS['__header_sent'] = 2;
	} // END - if
}

// Loads page footer and calls doShutdown()
function loadPageFooter () {
	// Init page footer
	$GLOBALS['__page_footer'] = '';

	// Footer disabled (e.g. CSS/AJAX/image output) or already sent?
	// 1234      5                         54    45                              5    5                              543    3                443    3                 443    3                  44321
	if ((((!isset($GLOBALS['__footer_sent'])) || (($GLOBALS['__footer_sent'] != 1) && ($GLOBALS['__footer_sent'] != 2))) && (!isCssOutputMode()) && (!isAjaxOutputMode()) && (!isImageOutputMode()))) {
		// Run the filter, sweet huh?
		runFilterChain('page_footer');

		// Load page footer
		$GLOBALS['__page_footer'] .= loadTemplate('page_footer', TRUE);
	} // END - if

	// Footer has been reached
	$GLOBALS['__footer_sent'] = 1;

	// Shutdown
	doShutdown();
}

// [EOF]
?>
